// Package circl provides a collection of cryptographic primitives. The goal
// of this module is to be used as a tool for experimental deployment of
// cryptographic algorithms targeting Post-Quantum (PQ) and Elliptic
// Curve Cryptography (ECC).
//
// Following blog post describes ideas behind CIRCL in more details:
// https://blog.cloudflare.com/introducing-circl/
package circl // github.com/cloudflare/circl
